# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from kolena._experimental.workflow.thresholded import ThresholdedMetrics
from kolena._utils.datatypes import get_args
from kolena._utils.datatypes import get_origin
from kolena.errors import WorkflowMismatchError
from kolena.workflow._datatypes import _SCALAR_TYPES
from kolena.workflow._datatypes import DATA_TYPE_FIELD
from kolena.workflow._datatypes import DataObject
from kolena.workflow.annotation import _ANNOTATION_TYPES
from kolena.workflow.asset import _ASSET_TYPES

_SUPPORTED_FIELD_TYPES = [*_SCALAR_TYPES, *_ANNOTATION_TYPES, *_ASSET_TYPES, ThresholdedMetrics]


def assert_workflows_match(workflow_expected: str, workflow_provided: str) -> None:
    if workflow_provided != workflow_expected:
        raise WorkflowMismatchError(
            f"workflow '{workflow_provided}' does not match expected workflow '{workflow_expected}'",
        )


def get_data_object_field_types(data_object_type: Type[DataObject]) -> Dict[str, Any]:
    # note that we use obj.__annotations__ instead of dataclasses.fields(obj) as the latter is not yet populated by the
    # time __init_subclass__ is called, blocking usage of this validator in __init_subclass__
    fields = getattr(data_object_type, "__annotations__", None)
    if fields is None:
        fields = {field.name: field.type for field in dataclasses.fields(data_object_type)}
    return fields


def validate_data_object_type(
    data_object_type: Type[DataObject],
    supported_field_types: Optional[List[Type]] = None,
    supported_list_types: Optional[List[Type]] = None,
) -> None:
    if not issubclass(data_object_type, DataObject):
        raise ValueError(f"'{data_object_type.__name__}' must extend {DataObject.__name__}")

    for field_name, field_type in get_data_object_field_types(data_object_type).items():
        validate_field(
            field_name,
            field_type,
            supported_field_types=supported_field_types,
            supported_list_types=supported_list_types,
        )


def validate_scalar_data_object_type(
    data_object_type: Type[DataObject],
    supported_list_types: Optional[List[Type]] = None,
) -> None:
    validate_data_object_type(
        data_object_type,
        supported_field_types=_SCALAR_TYPES,
        supported_list_types=supported_list_types or [],  # default to supporting no list types
    )


def validate_field(
    field_name: str,
    field_type: Type,
    supported_field_types: Optional[List[Type]] = None,
    supported_list_types: Optional[List[Type]] = None,
) -> None:
    if field_name == DATA_TYPE_FIELD:
        raise ValueError(f"Unsupported field name: '{DATA_TYPE_FIELD}' is reserved")

    supported_field_types = supported_field_types or _SUPPORTED_FIELD_TYPES
    supported_list_types = supported_field_types if supported_list_types is None else supported_list_types
    supported_bases = ", ".join(t.__name__ for t in supported_field_types)

    origin = get_origin(field_type)
    if origin is list:
        validate_list(field_name, field_type, supported_list_types)

    elif origin is Union:  # NOTE: get_origin(Optional[T]) == Union
        validate_union(field_name, field_type, supported_field_types, supported_list_types)

    elif origin is not None or field_type == typing.Any:
        raise ValueError(f"Unsupported field type: field '{field_name}', type '{field_type}'")

    elif not issubclass(field_type, tuple(supported_field_types)):
        raise ValueError(
            f"Unsupported field type: field '{field_name}', type '{field_type.__name__}'\n"
            f"Supported base types for fields: {supported_bases}",
        )


def validate_list(field_name: str, field_type: Type, supported_field_types: List[Type]) -> None:
    (arg,) = get_args(field_type)
    arg_origin = get_origin(arg)
    if arg_origin is Union:
        validate_union(field_name, arg, supported_field_types, supported_field_types)
    elif arg_origin is not None or not issubclass(arg, tuple(supported_field_types)):
        raise ValueError(
            f"Unsupported field type: field '{field_name}', unsupported List type '{arg}'",
        )


def validate_union(
    field_name: str,
    field_type: Union,
    supported_field_types: List[Type],
    supported_list_types: List[Type],
) -> None:
    args = get_args(field_type)
    type_name = "Optional" if type(None) in args and len(args) == 2 else "Union"
    for arg in args:
        err = f"Unsupported field type: field '{field_name}', unsupported {type_name} type '{arg}'"
        arg_origin = get_origin(arg)
        if arg_origin is list or arg_origin is List:
            validate_list(field_name, arg, supported_list_types)
            continue
        if arg_origin is not None or arg == typing.Any:  # e.g. Optional, Dict
            raise ValueError(err)
        if issubclass(arg, type(None)):
            continue
        if not issubclass(arg, tuple(supported_field_types)):
            raise ValueError(err)


def validate_metadata_dict(field_type: Type) -> None:
    origin = get_origin(field_type)
    if not (origin is dict or origin is Dict):
        raise ValueError("Metadata field must be dictionary type")
    args = get_args(field_type)
    if len(args) < 2:
        raise ValueError("Metadata dictionary must specify key and value types")
    if args[0] != str:
        raise ValueError(f"metadata.key type must be str: '{args[0]}'")
    validate_field("metadata.value", args[1], supported_field_types=[*_SCALAR_TYPES, type(None)])


def safe_issubclass(cls: Any, clsinfo: Any) -> bool:
    try:
        return issubclass(cls, clsinfo)
    except Exception:
        return False



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# noreorder
from kolena._experimental.dataset._dataset import fetch_dataset
from kolena._experimental.dataset._dataset import register_dataset
from kolena._experimental.dataset._evaluation import fetch_results
from kolena._experimental.dataset._evaluation import test

__all__ = [
    "register_dataset",
    "fetch_dataset",
    "fetch_results",
    "test",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
import os
import sys
from typing import Optional

import click

import kolena.workflow
from kolena._utils import repository
from kolena._utils.consts import KOLENA_TOKEN_ENV
from kolena._utils.state import _client_state
from kolena.errors import NotFoundError

KOLENA_COMMAND = "kolena"
CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])

_shared_options = [
    click.option(
        "--api-token",
        "-t",
        required=False,
        type=str,
        default=os.environ.get(KOLENA_TOKEN_ENV),
        help="api token for client to authorize to kolena server; if not specified, the command will use value of "
        f"environment variable '{KOLENA_TOKEN_ENV}'",
    ),
]

_shared_evaluator_options = [click.option("--workflow", "-w", required=True)]


def add_options(*args):
    def _add_options(func):
        options = [x for n in args for x in n]
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


@click.group(name=KOLENA_COMMAND, context_settings=CONTEXT_SETTINGS)
@click.option("--base-url", "-b", required=False, type=str, hidden=True, help="Use alternate sdk-server url")
def base_command(base_url: Optional[str] = None) -> None:
    if base_url:
        _client_state.base_url = base_url


@base_command.group(name="evaluator")
def evaluator_command() -> None:
    ...


@base_command.group(name="repository")
def repository_command() -> None:
    ...


@evaluator_command.command(name="register", help="Attach a remote evaluator to a particular workflow")
@add_options(_shared_options)
@add_options(_shared_evaluator_options)
@click.option(
    "--evaluator-name",
    "-n",
    required=True,
    help="name of the evaluator; must match evaluator name generated by image",
)
@click.option(
    "--image",
    "-i",
    required=True,
    help=(
        "name of the fully qualified image that has already been uploaded "
        "(eg: quay.io/kolena/pedestrian_crossing:0.1.0);"
        "NOTE: this is not validated at registration time but will cause issues at runtime "
        "if set to an invalid value"
    ),
)
@click.option(
    "--secret",
    "-s",
    required=False,
    help=(
        "sensitive data required by evaluator as string; "
        "NOTE: content would be store securely in kolena platform "
        "and accessible to the evaluator through environment variable"
    ),
)
@click.option(
    "--aws-assume-role",
    "-r",
    required=False,
    help="the AWS role arn the evaluator would assume to access AWS APIs",
)
def evaluator_register(
    workflow: str,
    evaluator_name: str,
    image: str,
    secret: Optional[str] = None,
    aws_assume_role: Optional[str] = None,
    api_token: Optional[str] = None,
) -> None:
    kolena.initialize(api_token=api_token)
    evaluator = kolena.workflow.workflow.register_evaluator(workflow, evaluator_name, image, secret, aws_assume_role)
    print(f"Image {image} successfully registered for workflow {workflow} evaluator {evaluator_name}.")
    print(json.dumps(dataclasses.asdict(evaluator), indent=2))


@evaluator_command.command(name="list", help="List registered remote evaluators for a particular workflow")
@add_options(_shared_options)
@add_options(_shared_evaluator_options)
def evaluator_list(workflow: str, api_token: Optional[str] = None) -> None:
    kolena.initialize(api_token=api_token)
    evaluators = kolena.workflow.workflow.list_evaluators(workflow)
    for evaluator in evaluators:
        print(f"evaluator_name='{evaluator.name}', image='{evaluator.image}', created='{evaluator.created}'")


@evaluator_command.command(name="get", help="Get the latest remote evaluator of a workflow")
@add_options(_shared_options)
@add_options(_shared_evaluator_options)
@click.option(
    "--evaluator-name",
    "-n",
    required=True,
    help="name of the evaluator",
)
@click.option(
    "--include-secret/--no-include-secret",
    "-s",
    help="include sensitive data registered with the evaluator or not",
    default=False,
)
def evaluator_get(workflow: str, evaluator_name: str, include_secret: bool, api_token: Optional[str] = None) -> None:
    kolena.initialize(api_token=api_token)
    try:
        evaluator = kolena.workflow.workflow.get_evaluator(workflow, evaluator_name, include_secret)
    except NotFoundError:
        print("Error: workflow not found", file=sys.stderr)
        sys.exit(1)

    print("Evaluator:")
    print(json.dumps(dataclasses.asdict(evaluator), indent=2))


@repository_command.command(name="create", help="Create a docker repository on kolena docker registry")
@add_options(_shared_options)
@click.option(
    "--name",
    "-n",
    required=True,
    help="repository name",
)
def repository_create(name: str, api_token: Optional[str] = None) -> None:
    kolena.initialize(api_token=api_token)
    try:
        repository.create(name)
    except Exception as e:
        print(f"Failed to create repository: {e}", file=sys.stderr)
        sys.exit(0)

    print("repository created")


def run():
    base_command()



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Optional
from typing import Type

from pydantic import validate_arguments

from kolena._api.v1.event import EventAPI
from kolena._utils import log
from kolena._utils.instrumentation import with_event
from kolena._utils.validators import ValidatorConfig
from kolena.detection import InferenceModel
from kolena.detection import Model
from kolena.detection import TestConfig
from kolena.detection import TestImage
from kolena.detection import TestSuite
from kolena.detection._datatypes import ImageResultDataFrame
from kolena.detection._internal.datatypes import LoadTestImagesDataFrame
from kolena.detection._internal.metadata import _from_dict
from kolena.detection._internal.test_run import BaseTestRun
from kolena.detection._internal.test_run import CustomMetricsCallback
from kolena.detection.inference import Inference
from kolena.detection.test_config import F1Optimal


class TestRun(BaseTestRun):
    """
    Interface to run tests for a [`Model`][kolena.detection.Model] on a [`TestSuite`][kolena.detection.TestSuite].

    For a streamlined interface, see [`test`][kolena.detection.test].

    Changes are committed to Kolena during execution and when the context is exited.

    :param model: The model being tested.
    :param test_suite: The test suite on which to test the model.
    :param test_config: Optionally specify a configuration, e.g.
        [`FixedGlobalThreshold`][kolena.detection.FixedGlobalThreshold] to customize the metrics evaluation logic for
        this test run. Defaults to [`F1Optimal`][kolena.detection.F1Optimal] with an `iou_threshold` of 0.5 if
        unspecified.
    :param custom_metrics_callback: Optionally specify a callback function to compute custom metrics for each test case.
        The callback would be passed inferences of images in each testcase and should return a dictionary with metric
        name as key and metric value as value.
    :param reset: Overwrites existing inferences if set.
    """

    _TestImageClass = TestImage
    _InferenceClass = Inference
    _ImageResultDataFrameClass: Type[ImageResultDataFrame] = ImageResultDataFrame
    _LoadTestImagesDataFrameClass: Type[LoadTestImagesDataFrame] = LoadTestImagesDataFrame

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        model: Model,
        test_suite: TestSuite,
        test_config: Optional[TestConfig] = None,
        custom_metrics_callback: Optional[CustomMetricsCallback[_TestImageClass, _InferenceClass]] = None,
        reset: bool = False,
    ):
        config = F1Optimal(iou_threshold=0.5) if test_config is None else test_config
        super().__init__(
            model,
            test_suite,
            config=config._to_run_config(),
            custom_metrics_callback=custom_metrics_callback,
            reset=reset,
        )

    def _image_from_load_image_record(self, record: Any) -> _TestImageClass:
        self._locator_to_image_id[record.locator] = record.test_sample_id
        return self._TestImageClass(
            locator=record.locator,
            dataset=record.dataset,
            metadata=_from_dict(record.metadata),
            ground_truths=[],  # note that no ground truths are loaded during testing
        )


@validate_arguments(config=ValidatorConfig)
@with_event(event_name=EventAPI.Event.EXECUTE_TEST_RUN)
def test(
    model: InferenceModel,
    test_suite: TestSuite,
    test_config: Optional[TestConfig] = None,
    custom_metrics_callback: Optional[CustomMetricsCallback[TestImage, Inference]] = None,
    reset: bool = False,
) -> None:
    """
    Test the provided [`InferenceModel`][kolena.detection.InferenceModel] on the provided
    [`TestSuite`][kolena.detection.TestSuite]. Any tests already in progress for this model on these suites are resumed.

    :param model: The model to test.
    :param test_suite: The test suite on which to test the model.
    :param test_config: Optionally specify a configuration, e.g.
        [`FixedGlobalThreshold`][kolena.detection.FixedGlobalThreshold], to customize metrics evaluation logic for this
        test run. Defaults to [`F1Optimal`][kolena.detection.F1Optimal] with an `iou_threshold` of 0.5 if unspecified.
    :param custom_metrics_callback: Optionally specify a callback function to compute custom metrics for each test case.
        The callback would be passed inferences of images in each testcase and should return a dictionary with metric
        name as key and metric value as value.
    :param reset: Overwrites existing inferences if set.
    """
    with TestRun(
        model,
        test_suite,
        test_config=test_config,
        custom_metrics_callback=custom_metrics_callback,
        reset=reset,
    ) as test_run:
        log.info("performing inference")
        for image in log.progress_bar(test_run.iter_images()):
            test_run.add_inferences(image, model.infer(image))
        log.success("performed inference")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noreorder
from kolena._api.v1.detection import CustomMetrics
from .test_config import TestConfig
from .test_config import FixedGlobalThreshold
from .test_config import AccuracyOptimal
from .test_image import TestImage
from .test_case import TestCase
from .test_suite import TestSuite
from .model import Model
from .model import InferenceModel
from .test_run import CustomMetricsCallback
from .test_run import TestRun
from .test_run import test

__all__ = [
    "TestConfig",
    "FixedGlobalThreshold",
    "AccuracyOptimal",
    "TestImage",
    "TestCase",
    "TestSuite",
    "Model",
    "InferenceModel",
    "CustomMetrics",
    "CustomMetricsCallback",
    "TestRun",
    "test",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import mimetypes
from dataclasses import asdict
from enum import Enum
from typing import Iterator
from typing import List
from typing import Optional
from typing import Union

import pandas as pd
import requests

from kolena._api.v2.dataset import EntityData
from kolena._api.v2.dataset import LoadDatapointsRequest
from kolena._api.v2.dataset import LoadDatasetByNameRequest
from kolena._api.v2.dataset import Path
from kolena._api.v2.dataset import RegisterRequest
from kolena._experimental.dataset.common import COL_DATAPOINT
from kolena._experimental.dataset.common import COL_DATAPOINT_ID_OBJECT
from kolena._experimental.dataset.common import validate_batch_size
from kolena._experimental.dataset.common import validate_dataframe_ids
from kolena._utils import krequests_v2 as krequests
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.serde import from_dict
from kolena._utils.state import API_V2
from kolena.errors import InputValidationError
from kolena.workflow._datatypes import _deserialize_dataobject
from kolena.workflow._datatypes import _serialize_dataobject
from kolena.workflow._datatypes import DATA_TYPE_FIELD
from kolena.workflow._datatypes import TypedDataObject
from kolena.workflow.io import _dataframe_object_serde

FIELD_LOCATOR = "locator"
FIELD_TEXT = "text"
SEP = "."


class DatapointType(str, Enum):
    AUDIO = "DATAPOINT/AUDIO"
    COMPOSITE = "DATAPOINT/COMPOSITE"
    DOCUMENT = "DATAPOINT/DOCUMENT"
    IMAGE = "DATAPOINT/IMAGE"
    POINT_CLOUD = "DATAPOINT/POINT_CLOUD"
    TABULAR = "DATAPOINT/TABULAR"
    TEXT = "DATAPOINT/TEXT"
    VIDEO = "DATAPOINT/VIDEO"

    @classmethod
    def has_value(cls, item) -> bool:
        return item in cls.__members__.values()


_DATAPOINT_TYPE_MAP = {
    "image": DatapointType.IMAGE.value,
    "application/pdf": DatapointType.DOCUMENT.value,
    "text": DatapointType.DOCUMENT.value,
    "video": DatapointType.VIDEO.value,
    "audio": DatapointType.AUDIO.value,
}


def _dataobject_type(obj: TypedDataObject) -> str:
    obj_type = obj._data_type()
    return f"{obj_type._data_category()}/{obj_type.value}"


def _get_datapoint_type(mimetype_str: str) -> str:
    main_type, sub_type = mimetype_str.split("/")
    return _DATAPOINT_TYPE_MAP.get(mimetype_str, None) or _DATAPOINT_TYPE_MAP.get(main_type, None)


def _infer_datatype_value(x: str) -> str:
    mtype, _ = mimetypes.guess_type(x)
    if mtype:
        datatype = _get_datapoint_type(mtype)
        if datatype is not None:
            return datatype
    elif x.endswith(".pcd"):
        return DatapointType.POINT_CLOUD.value

    return DatapointType.TABULAR.value


def _add_datatype(df: pd.DataFrame) -> None:
    """Adds `data_type` column(s) to input DataFrame."""
    prefixes = {
        column.rsplit(sep=SEP, maxsplit=1)[0]
        for column in df.columns.values
        if isinstance(column, str) and SEP in column
    }
    if prefixes:
        df[DATA_TYPE_FIELD] = DatapointType.COMPOSITE.value
        for prefix in prefixes:
            if not prefix.strip():
                raise InputValidationError(
                    "Empty prefix encountered when parsing composite dataset. "
                    f"Columns must lead with at least one non-whitespace character prior to delimeter '{SEP}'.",
                )
            if prefix in df.columns:
                raise InputValidationError(
                    f"Conflicting column '{prefix}' encountered when formatting composite dataset.",
                )
            if SEP in prefix:
                raise InputValidationError(
                    f"More than one delimeter '{SEP}' in prefix: '{prefix}'.",
                )

            composite_columns = df.filter(regex=rf"^{prefix}", axis=1).columns.to_list()
            composite = df.loc[:, composite_columns].rename(columns=lambda col: col.split(SEP)[-1])
            composite[DATA_TYPE_FIELD] = _infer_datatype(composite)

            df[prefix] = composite.to_dict("records")
            df.drop(columns=composite_columns, inplace=True)
    else:
        df[DATA_TYPE_FIELD] = _infer_datatype(df)


def _infer_datatype(df: pd.DataFrame) -> Union[pd.DataFrame, str]:
    if FIELD_LOCATOR in df.columns:
        return df[FIELD_LOCATOR].apply(_infer_datatype_value)
    elif FIELD_TEXT in df.columns:
        return DatapointType.TEXT.value

    return DatapointType.TABULAR.value


def _infer_id_fields(df: pd.DataFrame) -> List[str]:
    def get_id_fields_by(field: str) -> List[str]:
        return [
            id_field
            for id_field in df.columns.array
            if isinstance(id_field, str) and id_field.rsplit(SEP, maxsplit=1)[-1] == field
        ]

    if id_fields := get_id_fields_by(FIELD_LOCATOR):
        return id_fields
    elif id_fields := get_id_fields_by(FIELD_TEXT):
        return id_fields
    raise InputValidationError("Failed to infer the id_fields, please provide id_fields explicitly")


def _to_serialized_dataframe(df: pd.DataFrame, column: str) -> pd.DataFrame:
    result = _dataframe_object_serde(df, _serialize_dataobject)
    if column == COL_DATAPOINT:
        _add_datatype(result)
    result[column] = result.to_dict("records")
    result[column] = result[column].apply(lambda x: json.dumps(x))

    return result[[column]]


def _to_deserialized_dataframe(df: pd.DataFrame, column: str) -> pd.DataFrame:
    flattened = pd.json_normalize(
        [json.loads(r[column]) if r[column] is not None else {} for r in df.to_dict("records")],
        max_level=0,
    )
    flattened = _flatten_composite(flattened)
    flattened = flattened.loc[:, ~flattened.columns.str.endswith(DATA_TYPE_FIELD)]
    return _dataframe_object_serde(flattened, _deserialize_dataobject)


def _flatten_composite(df: pd.DataFrame) -> pd.DataFrame:
    for key, value in df.iloc[0].items():
        if isinstance(value, dict) and DatapointType.has_value(value.get(DATA_TYPE_FIELD)):
            flattened = pd.json_normalize(df[key], max_level=0).rename(
                columns=lambda col: f"{key}{SEP}{col}",
            )
            df = df.join(flattened)
            df.drop(columns=[key], inplace=True)
    return df


def _upload_dataset_chunk(df: pd.DataFrame, load_uuid: str, id_fields: List[str]) -> None:
    df_serialized_datapoint = _to_serialized_dataframe(df, column=COL_DATAPOINT)
    df_serialized_datapoint_id_object = _to_serialized_dataframe(df[sorted(id_fields)], column=COL_DATAPOINT_ID_OBJECT)
    df_serialized = pd.concat([df_serialized_datapoint, df_serialized_datapoint_id_object], axis=1)

    upload_data_frame(df=df_serialized, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=load_uuid)


def load_dataset(name: str) -> Optional[EntityData]:
    response = krequests.put(Path.LOAD_DATASET, json=asdict(LoadDatasetByNameRequest(name=name)))
    if response.status_code == requests.codes.not_found:
        return None

    response.raise_for_status()

    return from_dict(EntityData, response.json())


def resolve_id_fields(
    df: pd.DataFrame,
    id_fields: Optional[List[str]],
    existing_dataset: Optional[EntityData],
) -> List[str]:
    existing_id_fields = []
    if existing_dataset:
        existing_id_fields = existing_dataset.id_fields
    if not id_fields:
        if existing_id_fields:
            raise InputValidationError("id_fields is required for updating an existing dataset")
        else:
            id_fields = _infer_id_fields(df)
    return id_fields


def register_dataset(
    name: str,
    df: Union[Iterator[pd.DataFrame], pd.DataFrame],
    id_fields: Optional[List[str]] = None,
) -> None:
    """
    Create or update a dataset with datapoints and id_fields. If the dataset already exists, in order to associate the
    existing result with the new datapoints, the id_fields need be the same as the existing dataset.

    :param name: name of the dataset
    :param df: an iterator of pandas dataframe or a pandas dataframe, you can pass in the iterator if you want to have
                batch processing,
                 example iterator usage: csv_reader = pd.read_csv("PathToDataset.csv", chunksize=10)
    :param id_fields: a list of id fields, this will be used to link the result with the datapoints, if this is not
                 provided, it will be inferred from the dataset
    :return None
    """
    load_uuid = init_upload().uuid
    existing_dataset = load_dataset(name)
    if isinstance(df, pd.DataFrame):
        id_fields = resolve_id_fields(df, id_fields, existing_dataset)
        validate_dataframe_ids(df, id_fields)
        _upload_dataset_chunk(df, load_uuid, id_fields)
    else:
        validated = False
        for chunk in df:
            if not validated:
                id_fields = resolve_id_fields(chunk, id_fields, existing_dataset)
                validate_dataframe_ids(chunk, id_fields)
                validated = True
            _upload_dataset_chunk(chunk, load_uuid, id_fields)
    request = RegisterRequest(name=name, id_fields=id_fields, uuid=load_uuid)
    response = krequests.post(Path.REGISTER, json=asdict(request))
    krequests.raise_for_status(response)


def _iter_dataset_raw(
    name: str,
    batch_size: int = BatchSize.LOAD_SAMPLES.value,
) -> Iterator[pd.DataFrame]:
    validate_batch_size(batch_size)
    init_request = LoadDatapointsRequest(
        name=name,
        batch_size=batch_size,
    )
    yield from _BatchedLoader.iter_data(
        init_request=init_request,
        endpoint_path=Path.LOAD_DATAPOINTS.value,
        df_class=None,
        endpoint_api_version=API_V2,
    )


def iter_dataset(
    name: str,
    batch_size: int = BatchSize.LOAD_SAMPLES.value,
) -> Iterator[pd.DataFrame]:
    """
    Get an iterator over datapoints in the dataset.
    """
    for df_batch in _iter_dataset_raw(name, batch_size):
        yield _to_deserialized_dataframe(df_batch, column=COL_DATAPOINT)


def fetch_dataset(
    name: str,
    batch_size: int = BatchSize.LOAD_SAMPLES.value,
) -> pd.DataFrame:
    """
    Fetch an entire dataset given its name.
    """
    df_batches = list(iter_dataset(name, batch_size))
    return pd.concat(df_batches, ignore_index=True) if df_batches else pd.DataFrame()



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Assets are additional files linked to the [`TestSample`][kolena.workflow.TestSample],
[`GroundTruth`][kolena.workflow.GroundTruth], or [`Inference`][kolena.workflow.Inference] objects for your workflow.
Assets can be visualized in the Kolena Studio when exploring your test cases or model results.

The following asset types are available:

- [`ImageAsset`][kolena.workflow.asset.ImageAsset]
- [`PlainTextAsset`][kolena.workflow.asset.PlainTextAsset]
- [`BinaryAsset`][kolena.workflow.asset.BinaryAsset]
- [`PointCloudAsset`][kolena.workflow.asset.PointCloudAsset]
- [`VideoAsset`][kolena.workflow.asset.VideoAsset]

"""
from abc import ABCMeta
from typing import Optional

from pydantic.dataclasses import dataclass

from kolena._utils.validators import ValidatorConfig
from kolena.workflow._datatypes import _register_data_type
from kolena.workflow._datatypes import DataType
from kolena.workflow._datatypes import TypedDataObject


class _AssetType(DataType):
    IMAGE = "IMAGE"
    PLAIN_TEXT = "PLAIN_TEXT"
    BINARY = "BINARY"
    POINT_CLOUD = "POINT_CLOUD"
    VIDEO = "VIDEO"
    AUDIO = "AUDIO"

    @staticmethod
    def _data_category() -> str:
        return "ASSET"


@dataclass(frozen=True, config=ValidatorConfig)
class Asset(TypedDataObject[_AssetType], metaclass=ABCMeta):
    """Base class for all asset types."""

    def __init_subclass__(cls, **kwargs):
        _register_data_type(cls)


@dataclass(frozen=True, config=ValidatorConfig)
class ImageAsset(Asset):
    """An image in a cloud bucket."""

    locator: str
    """The location of this image in a cloud bucket, e.g. `s3://my-bucket/path/to/my-image-asset.png`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.IMAGE


@dataclass(frozen=True, config=ValidatorConfig)
class PlainTextAsset(Asset):
    """A plain text file in a cloud bucket."""

    locator: str
    """The location of this text file in a cloud bucket, e.g. `s3://my-bucket/path/to/my-text-asset.txt`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.PLAIN_TEXT


@dataclass(frozen=True, config=ValidatorConfig)
class BinaryAsset(Asset):
    """A binary file in a cloud bucket."""

    locator: str
    """The location of this text file in a cloud bucket, e.g. `s3://my-bucket/path/to/my-binary-asset.bin`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.BINARY


@dataclass(frozen=True, config=ValidatorConfig)
class PointCloudAsset(Asset):
    """
    A three-dimensional point cloud located in a cloud bucket. Points are assumed to be specified in a right-handed,
    Z-up coordinate system with the origin around the sensor that captured the point cloud.
    """

    locator: str
    """The location of this point cloud in a cloud bucket, e.g. `s3://my-bucket/path/to/my-point-cloud.pcd`."""

    @staticmethod
    def _data_type() -> _AssetType:
        return _AssetType.POINT_CLOUD


# NOTE: declare BaseVideoAsset as separate class for extension -- default fields in main VideoAsset class prevent
# extension with non-default fields
@dataclass(frozen=True, config=ValidatorConfig)
class BaseVideoAsset(Asset):
    """A video clip located in a cloud bucket or served at a URL."""

    locator: str
    """URL (e.g. S3, HTTPS) of the video file."""

    @classmethod
    def _data_type(cls) -> _AssetType:
        return _AssetType.VIDEO


@dataclass(frozen=True, config=ValidatorConfig)
class VideoAsset(BaseVideoAsset):
    """A video clip located in a cloud bucket or served at a URL."""

    locator: str
    """URL (e.g. S3, HTTPS) of the video file."""

    thumbnail: Optional[ImageAsset] = None
    """Optionally provide asset locator for custom video thumbnail image."""

    start: Optional[float] = None
    """Optionally specify start time of video snippet, in seconds."""

    end: Optional[float] = None
    """Optionally specify end time of video snippet, in seconds."""

    def __post_init__(self) -> None:
        if self.start is not None and self.end is not None and self.start > self.end:
            raise ValueError(f"Specified start time '{self.start}' is after specified end time '{self.end}'")
        if self.start is not None and self.end is not None and (self.start < 0 or self.end < 0):
            raise ValueError(f"Specified start time '{self.start}' and end time '{self.end}' must be non-negative")


_ASSET_TYPES = [ImageAsset, PlainTextAsset, BinaryAsset, PointCloudAsset, BaseVideoAsset, VideoAsset]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json
from abc import ABC
from collections import OrderedDict
from contextlib import contextmanager
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

import pandas as pd
from deprecation import deprecated
from pydantic import validate_arguments

from kolena._api.v1.event import EventAPI
from kolena._api.v1.fr import TestCase as API
from kolena._utils import krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.consts import FieldName
from kolena._utils.dataframes.validators import validate_df_schema
from kolena._utils.frozen import Frozen
from kolena._utils.instrumentation import with_event
from kolena._utils.instrumentation import WithTelemetry
from kolena._utils.serde import from_dict
from kolena._utils.validators import validate_name
from kolena._utils.validators import ValidatorConfig
from kolena.errors import NotFoundError
from kolena.fr.datatypes import TEST_CASE_COLUMNS
from kolena.fr.datatypes import TestCaseDataFrame
from kolena.fr.datatypes import TestCaseDataFrameSchema
from kolena.fr.datatypes import TestCaseRecord


class TestCase(ABC, Frozen, WithTelemetry):
    """
    A group of image pairs that can be added to a [`TestSuite`][kolena.fr.TestSuite].

    The test case is the base unit of results computation in the Kolena platform. Metrics are computed by test case.
    """

    name: str
    """The unique name of this test case. Cannot be changed after creation."""

    version: int
    """
    The version of this test case. A test case's version is automatically incremented whenever it is edited via
    [`TestCase.edit`][kolena.fr.TestCase.edit].
    """

    description: str
    """
    Free-form, human-readable description of this test case. Can be edited at any time via
    [`TestCase.edit`][kolena.fr.TestCase.edit].
    """

    image_count: int
    """The number of images included in this test case."""

    pair_count_genuine: int
    """The number of genuine image pairs included in this test case."""

    pair_count_imposter: int
    """The number of imposter image pairs included in this test case."""

    Data = API.EntityData

    _id: int
    _data: API.EntityData

    def __init__(
        self,
        name: str,
        version: Optional[int] = None,
        description: Optional[str] = None,
        test_samples: Optional[List[TestCaseRecord]] = None,
        reset: bool = False,
    ):
        validate_name(name, FieldName.TEST_CASE_NAME)
        try:
            self._populate_from_other(self.load(name, version))
            if description is not None and self.description != description and not reset:
                log.warn("test case already exists, not updating description when reset=False")
            if test_samples is not None:
                if self.version > 0 and not reset:
                    log.warn("not updating test samples for test case that has already been edited when reset=False")
                else:
                    self._hydrate(test_samples, description)
        except NotFoundError:
            if version is not None:
                log.warn(f"creating new test case '{name}', ignoring provided version")
            self._populate_from_other(self.create(name, description, test_samples))
        self._freeze()

    @property
    @deprecated(details="use `TestCase` instance attributes", deprecated_in="0.57.0")
    def data(self) -> API.EntityData:
        """
        !!! warning "Deprecated: since `0.57.0`"
            Access this data via instance attributes, e.g. `self.image_count`, directly.

        The data associated with this test case.
        """
        return self._data

    @data.setter
    @deprecated(details="use `TestCase` instance attributes", deprecated_in="0.57.0")
    def data(self, new_data: API.EntityData) -> None:
        self._data = new_data

    @classmethod
    @with_event(event_name=EventAPI.Event.CREATE_TEST_CASE)
    def create(
        cls,
        name: str,
        description: Optional[str] = None,
        test_samples: Optional[List[TestCaseRecord]] = None,
    ) -> "TestCase":
        """
        Create a new test case with the provided name.

        :param name: The name of the new test case to create.
        :param description: Optional free-form description of the test case to create.
        :param test_samples: Optionally specify a set of test samples to populate the test case.
        :return: The newly created test case.
        """
        validate_name(name, FieldName.TEST_CASE_NAME)
        request = API.CreateRequest(name=name, description=description or "")
        res = krequests.post(endpoint_path=API.Path.CREATE.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=API.EntityData, data=res.json())
        obj = cls._create_from_data(data)
        log.info(f"created test case '{name}' (v{obj.version})")
        if test_samples is not None:
            obj._hydrate(test_samples)
        return obj

    @classmethod
    @with_event(event_name=EventAPI.Event.LOAD_TEST_CASE)
    def load(cls, name: str, version: Optional[int] = None) -> "TestCase":
        """
        Load an existing test case with the provided name.

        :param name: The name of the test case to load.
        :param version: Optionally specify a particular version of the test case to load. Defaults to the latest version
            when unset.
        :return: The loaded test case.
        """
        return cls._load_by_name(name, version)

    @classmethod
    @deprecated(details="use `TestCase.load`", deprecated_in="0.57.0")
    def load_by_name(cls, name: str, version: Optional[int] = None) -> "TestCase":
        """
        !!! warning "Deprecated: since `0.57.0`"
            Use [`TestCase.load`][kolena.fr.TestCase.load] instead.

        Load an existing test case with the provided name.

        :param name: The name of the test case to load.
        :param version: optionally specify the target version of the test case to load. When absent, the highest version
            of the test case with the provided name is returned
        :return: the loaded test case
        """
        return cls.load(name, version)

    @classmethod
    def _load_by_name(cls, name: str, version: Optional[int] = None) -> "TestCase":
        request = API.LoadByNameRequest(name=name, version=version)
        res = krequests.put(endpoint_path=API.Path.LOAD_BY_NAME.value, data=json.dumps(dataclasses.asdict(request)))
        krequests.raise_for_status(res)
        data = from_dict(data_class=API.EntityData, data=res.json())
        log.info(f"loaded test case '{name}' (v{data.version})")
        return cls._create_from_data(data)

    def load_data(self) -> TestCaseDataFrame:
        """
        Load all image pairs for a test case.

        :return: DataFrame containing all pairs defined in this test case.
        """
        return _BatchedLoader.concat(self.iter_data(), TestCaseDataFrame)

    @classmethod
    def _create_from_data(cls, data: API.EntityData) -> "TestCase":
        obj = cls.__new__(cls)
        obj._id = data.id
        obj.name = data.name
        obj.version = data.version
        obj.description = data.description
        obj.image_count = data.image_count
        obj.pair_count_genuine = data.pair_count_genuine
        obj.pair_count_imposter = data.pair_count_imposter
        obj.data = data
        obj._freeze()
        return obj

    def _hydrate(self, test_samples: List[TestCaseRecord], description: Optional[str] = None) -> None:
        if len(test_samples) == 0:
            log.warn("no test samples provided, unable to populate test case")
            return
        with self.edit(reset=True) as editor:
            if description is not None:
                editor.description(description)
            for locator_a, locator_b, is_same in test_samples:
                editor.add(locator_a, locator_b, is_same)

    def _populate_from_other(self, other: "TestCase") -> None:
        with self._unfrozen():
            self._id = other._id
            self.name = other.name
            self.version = other.version
            self.description = other.description
            self.image_count = other.image_count
            self.pair_count_genuine = other.pair_count_genuine
            self.pair_count_imposter = other.pair_count_imposter
            self.data = other._data

    class Editor:
        _samples: Dict[str, TestCaseRecord]
        _reset: bool
        _description: str
        _initial_description: str
        _initial_samples: Optional[Dict[str, TestCaseRecord]] = None

        def __init__(self, description: str, reset: bool = False) -> None:
            self._reset = reset
            self._description = description
            self._initial_description = description
            self._samples: Dict[str, TestCaseRecord] = OrderedDict()

        @validate_arguments(config=ValidatorConfig)
        def description(self, description: str) -> None:
            """
            Update the description of this test case.

            :param description: The new test case description.
            """
            self._description = description

        @validate_arguments(config=ValidatorConfig)
        def add(self, locator_a: str, locator_b: str, is_same: bool) -> None:
            """
            Add the provided image pair to the test case.

            Note that if the image pair with `locator_a` and `locator_b` is already defined within the platform,
            the value for `is_same` must match the value already defined.

            :param locator_a: The left locator for the image pair.
            :param locator_b: The right locator for the image pair.
            :param is_same: Whether to treat this image pair as a a genuine pair (`True`) or an imposter pair (`False`).
            """
            key = self._key(locator_a, locator_b)
            val = (locator_a, locator_b, is_same)
            if val == self._samples.get(key, None):
                log.info(f"no op: {val} already in test case")
                return
            self._samples[key] = val

        @validate_arguments(config=ValidatorConfig)
        def remove(self, locator_a: str, locator_b: str) -> None:
            """
            Remove the provided pair from the test case.

            :param locator_a: The left locator for the image pair.
            :param locator_b: The right locator for the image pair.
            :raises KeyError: If the provided locator pair is not in the test case.
            """
            key = self._key(locator_a, locator_b)
            if key not in self._samples.keys():
                raise KeyError(f"pair not in test case: {locator_a}, {locator_b}")
            self._samples.pop(key)

        @staticmethod
        def _key(locator_a: str, locator_b: str) -> str:
            # newline is guaranteed to not be present in the locators
            return f"{locator_a}\n{locator_b}"

        def _edited(self) -> bool:
            return (
                self._reset or self._description != self._initial_description or self._samples != self._initial_samples
            )

    @contextmanager
    @with_event(event_name=EventAPI.Event.EDIT_TEST_CASE)
    def edit(self, reset: bool = False) -> Iterator[Editor]:
        """
        Edit this test case in a context:

        ```python
        with test_case.edit() as editor:
            # perform as many editing actions as desired
            editor.add(...)
            editor.remove(...)
        ```

        Changes are committed to the Kolena platform when the context is exited.

        :param reset: Clear all existing test samples in the test case.
        """
        editor = self.Editor(self.description, reset)

        if not reset:
            df_existing = self.load_data()
            # avoid calling the expensive self.load_data() multiple times
            _initial_samples: Dict[str, TestCaseRecord] = OrderedDict()
            for record in df_existing.itertuples():
                locator_a, locator_b, is_same = record.locator_a, record.locator_b, record.is_same
                test_case_record = (locator_a, locator_b, is_same)
                editor.add(*test_case_record)
                _initial_samples[editor._key(locator_a, locator_b)] = test_case_record
            editor._initial_samples = _initial_samples

        yield editor

        # no-op contexts have no effect, do not bump version
        if not editor._edited():
            return

        log.info(f"editing test case '{self.name}' (v{self.version})")
        init_response = init_upload()
        df = pd.DataFrame(editor._samples.values(), columns=TEST_CASE_COLUMNS)
        df_validated = validate_df_schema(df, TestCaseDataFrameSchema)

        upload_data_frame(df=df_validated, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=init_response.uuid)
        request = API.CompleteEditRequest(
            test_case_id=self._id,
            current_version=self.version,
            name=self.name,
            description=editor._description,
            uuid=init_response.uuid,
        )
        complete_res = krequests.post(
            endpoint_path=API.Path.COMPLETE_EDIT.value,
            data=json.dumps(dataclasses.asdict(request)),
        )
        krequests.raise_for_status(complete_res)
        test_case_data = from_dict(data_class=API.EntityData, data=complete_res.json())
        self._populate_from_other(self._create_from_data(test_case_data))
        log.success(f"edited test case '{self.name}' (v{self.version})")

    @validate_arguments
    def iter_data(self, batch_size: int = 10_000_000) -> Iterator[TestCaseDataFrame]:
        """
        Iterator of DataFrames describing all pairs data for a test case.

        :param batch_size: Optionally specify maximum number of rows to be returned in a single DataFrame.
        """
        log.info(f"loading image pairs in test case '{self.name}' (v{self.version})")
        init_request = API.InitLoadDataRequest(batch_size=batch_size, test_case_id=self._id)
        yield from _BatchedLoader.iter_data(
            init_request=init_request,
            endpoint_path=API.Path.INIT_LOAD_DATA.value,
            df_class=TestCaseDataFrame,
        )
        log.info(f"loaded image pairs in test case '{self.name}' (v{self.version})")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from kolena._api.v1.workflow import WorkflowType
from kolena.detection import Inference
from kolena.detection import TestCase
from kolena.detection import TestImage
from kolena.detection import TestSuite
from kolena.detection._datatypes import LoadInferencesDataFrame
from kolena.detection._internal import BaseModel
from kolena.detection.inference import BoundingBox
from kolena.detection.inference import SegmentationMask


class Model(BaseModel):
    """
    The descriptor for an object detection model within Kolena.

    For additional functionality, see the associated
    [base class documentation][kolena.detection._internal.model.BaseModel].
    """

    name: str
    """
    Unique name of the model within the platform. If the provided model name has already been registered, that model
    and its metadata are loaded upon instantiation.
    """

    metadata: Dict[str, Any]
    """Unstructured metadata associated with the model."""

    _TestImageClass = TestImage
    _TestCaseClass = TestCase
    _TestSuiteClass = TestSuite
    _InferenceClass = BoundingBox
    _LoadInferencesDataFrameClass = LoadInferencesDataFrame

    def __init__(self, name: str, metadata: Optional[Dict[str, Any]] = None):
        super().__init__(name, WorkflowType.DETECTION, metadata)

    def _inferences_from_record(self, record: Any) -> Tuple[TestImage, Optional[List[_InferenceClass]]]:
        test_image = TestImage._from_record(record)
        if record.inferences is None:
            return test_image, None
        inferences = [Inference._from_dict(d) for d in record.inferences]
        bboxes = [inference for inference in inferences if isinstance(inference, (BoundingBox, SegmentationMask))]
        return test_image, bboxes


class InferenceModel(Model):
    """
    Extension of [`Model`][kolena.detection.Model] with an `infer` method to perform inference.

    See documentation parent [`Model`][kolena.detection.Model] for details.
    """

    infer: Callable[[TestImage], Optional[List[Inference]]]
    """
    Function transforming a [`TestImage`][kolena.detection.TestImage] into a list of zero or more
    [`Inference`][kolena.detection.Inference] objects.
    """

    def __init__(
        self,
        name: str,
        infer: Callable[[TestImage], Optional[List[Inference]]],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        setattr(self, "infer", infer)  # bypass mypy method assignment bug
        super().__init__(name, metadata)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Metadata associated with a [`TestImage`][kolena.classification.TestImage].

```python
from kolena.classification import TestImage
from kolena.classification.metadata import Landmarks, BoundingBox, Asset

test_image = TestImage("s3://bucket/path/to/image.png", metadata=dict(
    input_landmarks=Landmarks([(0,0), (10, 10), (20, 20), (30, 30), (40, 40)]),
    input_bounding_box=BoundingBox((0, 0), (100, 100)),
    image_grayscale=Asset("s3://bucket/path/to/image_grayscale.png"),
))
```
"""
from kolena.detection._internal.metadata import Annotation
from kolena.detection._internal.metadata import Asset
from kolena.detection._internal.metadata import BoundingBox
from kolena.detection._internal.metadata import Landmarks
from kolena.detection._internal.metadata import MetadataElement

__all__ = [
    "Annotation",
    "BoundingBox",
    "Landmarks",
    "Asset",
    "MetadataElement",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .evaluator import InstanceSegmentationEvaluator
from .workflow import EvaluatorConfiguration
from .workflow import GroundTruth
from .workflow import Inference
from .workflow import Model
from .workflow import TestCase
from .workflow import TestSample
from .workflow import TestSuite

__all__ = [
    "EvaluatorConfiguration",
    "GroundTruth",
    "Inference",
    "Model",
    "TestSample",
    "TestCase",
    "TestSuite",
    "InstanceSegmentationEvaluator",
]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import List

from pydantic.dataclasses import dataclass

from kolena._api.v1.batched_load import BatchedLoad


class Path(str, Enum):
    REGISTER = "/dataset/register"
    LOAD_DATAPOINTS = "/dataset/load-datapoints"
    LOAD_DATASET = "/dataset/load-by-name"


@dataclass(frozen=True)
class RegisterRequest:
    name: str
    id_fields: List[str]
    uuid: str


@dataclass(frozen=True)
class LoadDatapointsRequest(BatchedLoad.BaseInitDownloadRequest):
    name: str


@dataclass(frozen=True)
class LoadDatasetByNameRequest:
    name: str


@dataclass(frozen=True)
class EntityData:
    id: int
    name: str
    description: str
    id_fields: List[str]



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import cast
from typing import Generic
from typing import Optional
from typing import Type
from typing import TypeVar

import pandas as pd
import pandera as pa

from kolena._utils.dataframes.validators import validate_df_schema

# export these such that this version handling logic only needs to be applied here
try:
    # Python >= 3.8
    from typing import get_args
    from typing import get_origin
except ImportError:

    def get_args(t: Any) -> tuple:
        return getattr(t, "__args__", ())

    def get_origin(t: Any) -> Optional[Type]:
        return getattr(t, "__origin__", None)


TDataFrame = TypeVar("TDataFrame", bound="LoadableDataFrame")
TSchema = TypeVar("TSchema", bound=pa.SchemaModel)


class LoadableDataFrame(ABC, Generic[TDataFrame]):
    @classmethod
    @abstractmethod
    def get_schema(cls) -> Type[TSchema]:
        raise NotImplementedError

    @classmethod
    def construct_empty(cls) -> TDataFrame:
        df = pd.DataFrame({key: [] for key in cls.get_schema().to_schema().columns.keys()})
        return cast(TDataFrame, validate_df_schema(df, cls.get_schema(), trusted=True))

    @classmethod
    def from_serializable(cls, df: pd.DataFrame) -> TDataFrame:
        return cast(TDataFrame, validate_df_schema(df, cls.get_schema(), trusted=True))



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from pydantic import validate_arguments

from kolena._api.v1.workflow import WorkflowType
from kolena._utils.validators import ValidatorConfig
from kolena.classification import TestCase
from kolena.classification import TestImage
from kolena.classification import TestSuite
from kolena.detection import Inference
from kolena.detection._datatypes import LoadInferencesDataFrame
from kolena.detection._internal import BaseModel
from kolena.detection.inference import ClassificationLabel


class Model(BaseModel):
    """
    The descriptor for a classification model in Kolena.

    For additional functionality, see the associated
    [base class documentation][kolena.detection._internal.model.BaseModel].
    """

    name: str
    """
    Unique name of the model, potentially containing information about the architecture, training dataset,
    configuration, framework, commit hash, etc.
    """

    metadata: Dict[str, Any]
    """Unstructured metadata associated with the model."""

    _TestImageClass = TestImage
    _TestCaseClass = TestCase
    _TestSuiteClass = TestSuite
    _InferenceClass = Tuple[str, float]
    _LoadInferencesDataFrameClass = LoadInferencesDataFrame

    @validate_arguments(config=ValidatorConfig)
    def __init__(self, name: str, metadata: Optional[Dict[str, Any]] = None):
        super().__init__(name, WorkflowType.CLASSIFICATION, metadata)

    def _inferences_from_record(self, record: Any) -> Tuple[TestImage, Optional[List[_InferenceClass]]]:
        test_image = TestImage._from_record(record)
        if record.inferences is None:
            return test_image, None
        detection_inferences = [Inference._from_dict(d) for d in record.inferences]
        inferences = [
            (inference.label, inference.confidence)
            for inference in detection_inferences
            if isinstance(inference, ClassificationLabel)
        ]
        return test_image, inferences


class InferenceModel(Model):
    """
    A [`Model`][kolena.classification.Model] with a special [`infer`][kolena.classification.InferenceModel.infer] member
    performing inference on a provided [`TestImage`][kolena.classification.TestImage].
    """

    infer: Callable[[TestImage], Optional[List[Tuple[str, float]]]]
    """
    A function transforming an input [`TestImage`][kolena.classification.TestImage] to zero or more
    `(label, confidence)` tuples representing model predictions.
    """

    @validate_arguments(config=ValidatorConfig)
    def __init__(
        self,
        name: str,
        infer: Callable[[TestImage], Optional[List[Tuple[str, float]]]],
        metadata: Optional[Dict[str, Any]] = None,
    ):
        setattr(self, "infer", infer)  # bypass mypy method assignment bug
        super().__init__(name, metadata)



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from dataclasses import asdict
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pandas as pd

from kolena._api.v2.model import LoadResultsRequest
from kolena._api.v2.model import Path
from kolena._api.v2.model import UploadResultsRequest
from kolena._experimental.dataset._dataset import _iter_dataset_raw
from kolena._experimental.dataset._dataset import _to_deserialized_dataframe
from kolena._experimental.dataset._dataset import _to_serialized_dataframe
from kolena._experimental.dataset._dataset import load_dataset
from kolena._experimental.dataset.common import COL_DATAPOINT
from kolena._experimental.dataset.common import COL_DATAPOINT_ID_OBJECT
from kolena._experimental.dataset.common import COL_EVAL_CONFIG
from kolena._experimental.dataset.common import COL_RESULT
from kolena._experimental.dataset.common import validate_batch_size
from kolena._experimental.dataset.common import validate_dataframe_ids
from kolena._utils import krequests_v2 as krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.state import API_V2
from kolena.errors import IncorrectUsageError
from kolena.errors import NotFoundError

TYPE_EVALUATION_CONFIG = Optional[Dict[str, Any]]
TEST_ON_TYPE = Optional[Union[str, List[str]]]


def _fetch_dataset(dataset: str) -> pd.DataFrame:
    df_data_batch = list(_iter_dataset_raw(dataset))
    df_datapoints = pd.concat(df_data_batch) if df_data_batch else pd.DataFrame(columns=["id", COL_DATAPOINT])
    df_datapoints.rename(columns={"id": "datapoint_id"}, inplace=True)
    return df_datapoints


def _iter_result_raw(dataset: str, model: str, batch_size: int) -> Iterator[pd.DataFrame]:
    validate_batch_size(batch_size)
    init_request = LoadResultsRequest(dataset=dataset, model=model, batch_size=batch_size)
    yield from _BatchedLoader.iter_data(
        init_request=init_request,
        endpoint_path=Path.LOAD_RESULTS.value,
        df_class=None,
        endpoint_api_version=API_V2,
    )


def _fetch_results(dataset: str, model: str) -> pd.DataFrame:
    df_result_batch = list(_iter_result_raw(dataset, model, batch_size=BatchSize.LOAD_RECORDS))
    return (
        pd.concat(df_result_batch)
        if df_result_batch
        else pd.DataFrame(
            columns=["datapoint_id", COL_DATAPOINT, COL_RESULT, COL_EVAL_CONFIG],
        )
    )


def _process_result(
    eval_config: Optional[TYPE_EVALUATION_CONFIG],
    df_result: pd.DataFrame,
    id_fields: List[str],
) -> pd.DataFrame:
    df_serialized_datapoint_id_object = _to_serialized_dataframe(
        df_result[sorted(id_fields)],
        column=COL_DATAPOINT_ID_OBJECT,
    )
    df_result_eval = _to_serialized_dataframe(df_result.drop(columns=id_fields), column=COL_RESULT)
    df_result_eval[COL_EVAL_CONFIG] = json.dumps(eval_config) if eval_config is not None else None
    df_result_eval = pd.concat([df_result_eval, df_serialized_datapoint_id_object], axis=1)
    return df_result_eval


def _upload_results(
    model: str,
    load_uuid: str,
    dataset_id: int,
) -> None:
    request = UploadResultsRequest(
        model=model,
        uuid=load_uuid,
        dataset_id=dataset_id,
    )
    response = krequests.post(Path.UPLOAD_RESULTS, json=asdict(request))
    krequests.raise_for_status(response)


def fetch_results(
    dataset: str,
    model: str,
) -> Tuple[pd.DataFrame, List[Tuple[TYPE_EVALUATION_CONFIG, pd.DataFrame]]]:
    """
    Fetch results given dataset name and model name.
    """
    df = _fetch_results(dataset, model)

    df_datapoints = _to_deserialized_dataframe(df.drop_duplicates(subset=[COL_DATAPOINT]), column=COL_DATAPOINT)
    eval_configs = df[COL_EVAL_CONFIG].unique()
    df_results_by_eval = []
    for eval_config in eval_configs:
        df_matched = df[df[COL_EVAL_CONFIG] == eval_config if eval_config is not None else df[COL_EVAL_CONFIG].isnull()]
        df_results_by_eval.append(
            (
                json.loads(eval_config) if eval_config is not None else None,
                _to_deserialized_dataframe(df_matched, column=COL_RESULT),
            ),
        )

    return df_datapoints, df_results_by_eval


def _validate_configs(configs: List[TYPE_EVALUATION_CONFIG]) -> None:
    n = len(configs)
    for i in range(n):
        for j in range(i + 1, n):
            if configs[i] == configs[j]:
                raise IncorrectUsageError("duplicate eval configs are invalid")


def test(
    dataset: str,
    model: str,
    results: Union[
        pd.DataFrame,
        Iterator[pd.DataFrame],
        List[
            Tuple[
                TYPE_EVALUATION_CONFIG,
                Union[
                    pd.DataFrame,
                    Iterator[pd.DataFrame],
                ],
            ]
        ],
    ],
) -> None:
    """
    This function is used for testing a specified model on a given dataset.

    :param dataset: The name of the dataset to be used.
    :param model: The name of the model to be used.
    :param results: Either a DataFrame or a list of tuples, where each tuple consists of
                    a eval configuration and a DataFrame.
    :return None
    """
    existing_dataset = load_dataset(dataset)
    if not existing_dataset:
        raise NotFoundError(f"dataset {dataset} does not exist")

    if isinstance(results, pd.DataFrame) or isinstance(results, Iterator):
        results = [(None, results)]
    load_uuid = init_upload().uuid

    _validate_configs([cfg for cfg, _ in results])
    for config, df_result_input in results:
        log.info(f"start evaluation with configuration {config}" if config else "start evaluation")
        if isinstance(df_result_input, pd.DataFrame):
            validate_dataframe_ids(df_result_input, existing_dataset.id_fields)
            df_results = _process_result(config, df_result_input, existing_dataset.id_fields)
            upload_data_frame(df=df_results, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=load_uuid)
        else:
            id_column_validated = False
            for df_result in df_result_input:
                if not id_column_validated:
                    validate_dataframe_ids(df_result, existing_dataset.id_fields)
                    id_column_validated = True
                df_results = _process_result(config, df_result, existing_dataset.id_fields)
                upload_data_frame(df=df_results, batch_size=BatchSize.UPLOAD_RECORDS.value, load_uuid=load_uuid)

    _upload_results(model, load_uuid, existing_dataset.id)
    log.info(f"uploaded test results for model {model} on dataset {dataset}")



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Optional

import numpy as np

from kolena.errors import InputValidationError


class Colormap(ABC):
    """
    A `Colormap` maps a pixel intensity to RGBA.
    """

    fade_low_activation: bool
    """
    Fades out the regions with low activation by applying zero alpha value if set
    `True`; otherwise, activation map is shown as is without any fading applied. By default, it's set to `True`.
    This option makes the overlay visualization better by highlighting only the important regions.
    """

    def __init__(self, fade_low_activation: Optional[bool] = True):
        self.fade_low_activation = fade_low_activation

    @abstractmethod
    def red(self, intensity: np.uint8) -> np.uint8:
        """
        Maps a grayscale pixel intensity to color red: [0, 255]
        """
        raise NotImplementedError

    @abstractmethod
    def green(self, intensity: np.uint8) -> np.uint8:
        """
        Maps a grayscale pixel intensity to color green: [0, 255]
        """
        raise NotImplementedError

    @abstractmethod
    def blue(self, intensity: np.uint8) -> np.uint8:
        """
        Maps a grayscale pixel intensity to color blue: [0, 255]
        """
        raise NotImplementedError

    def alpha(self, intensity: np.uint8) -> np.uint8:
        """
        Maps the grayscale pixel intensity to alpha: [0, 255]. If `fade_low_activation`
        is False, then it returns the maximum alpha value.
        """
        max_uint8 = np.iinfo(np.uint8).max
        if self.fade_low_activation:
            # apply non-linear scaling alpha channel [0, 255]
            return max_uint8 / (1 + np.exp(((max_uint8 / 2) - intensity) / 8))
        else:
            return max_uint8

    def colorize(self, intensity: np.uint8) -> np.array:
        return np.array(
            [
                self.red(intensity),
                self.green(intensity),
                self.blue(intensity),
                self.alpha(intensity),
            ],
            dtype=np.uint8,
        )

    def _interpolate(self, intensity: np.uint8, x0: float, y0: float, x1: float, y1: float) -> np.uint8:
        return np.uint8(round((intensity - x0) * (y1 - y0) / (x1 - x0) + y0))


class ColormapJet(Colormap):
    """
    The [MATLAB "Jet" color palette](http://blogs.mathworks.com/images/loren/73/colormapManip_14.png) is a standard
    palette used for scientific and mathematical data.

    It is defined as a linear ramp between the following colours: "#00007F", "blue", "#007FFF", "cyan", "#7FFF7F",
    "yellow", "#FF7F00", "red", "#7F0000"
    """

    def _scale_to_colormap(self, intensity: np.uint8) -> np.uint8:
        max_uint8 = np.iinfo(np.uint8).max
        if intensity <= max_uint8 / 8:
            return 0
        elif intensity <= max_uint8 * 3 / 8:
            return self._interpolate(intensity, max_uint8 / 8, 0.0, max_uint8 * 3 / 8, max_uint8)
        elif intensity <= max_uint8 * 5 / 8:
            return max_uint8
        elif intensity <= max_uint8 * 7 / 8:
            return self._interpolate(intensity, max_uint8 * 5 / 8, max_uint8, max_uint8 * 7 / 8, 0.0)
        else:
            return 0

    def red(self, intensity: np.uint8) -> np.uint8:
        return self._scale_to_colormap(intensity - np.iinfo(np.uint8).max / 4)

    def green(self, intensity: np.uint8) -> np.uint8:
        return self._scale_to_colormap(intensity)

    def blue(self, intensity: np.uint8) -> np.uint8:
        return self._scale_to_colormap(intensity + np.iinfo(np.uint8).max / 4)


def colorize_activation_map(activation_map: np.ndarray, colormap: Optional[Colormap] = ColormapJet()) -> np.ndarray:
    """
    Applies the specified colormap to the activation map.

    :param activation_map: A 2D numpy array, shaped (h, w) or (h, w, 1), of the activation map in `np.uint8`
        or `float` ranging [0, 1].
    :param colormap: The colormap used to colorize the input activation map. Defaults to the
        [MATLAB "Jet" colormap](http://blogs.mathworks.com/images/loren/73/colormapManip_14.png).
    :return: The colorized activation map in RGBA format, in (h, w, 4) shape.
    """
    max_uint8 = np.iinfo(np.uint8).max

    if activation_map.size == 0:
        raise InputValidationError("input array is empty")

    if np.issubdtype(activation_map.dtype, np.floating) and ((0 <= activation_map) & (activation_map <= 1)).all():
        activation_map = np.rint(activation_map * max_uint8).astype(np.uint8)

    if activation_map.dtype != np.uint8:
        raise InputValidationError(f"input array type must be np.uint8, but received {activation_map.dtype}")

    if len(activation_map.shape) != 2 and len(activation_map.shape) != 3:
        raise InputValidationError(
            f"input array must have 2 or 3 dimensions, but received {len(activation_map.shape)}\n"
            "Expected shape is (h, w) or (h, w, 1)",
        )

    if len(activation_map.shape) == 2:
        activation_map = np.expand_dims(activation_map, axis=2)

    if len(activation_map.shape) == 3:
        if activation_map.shape[2] != 1:
            raise InputValidationError(
                f"input image must be a single-channel, but received {activation_map.shape[2]} channels\n"
                "Expected shape is (h, w) or (h, w, 1)",
            )
        activation_map = np.repeat(activation_map, 4, axis=2)

    vcolorize = np.vectorize(colormap.colorize, signature="()->(n)")
    activation_map[:, :] = vcolorize(activation_map[:, :, 0])

    return activation_map



# Copyright 2021-2023 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from pydantic.dataclasses import dataclass

from kolena._api.v1.batched_load import BatchedLoad


class Model:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        metadata: Dict[str, Any]
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str

    @dataclass(frozen=True)
    class LoadAllRequest:
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        metadata: Dict[str, Any]
        tags: Set[str]
        workflow: str

    @dataclass(frozen=True)
    class LoadAllResponse:
        models: List["Model.EntityData"]

    @dataclass(frozen=True)
    class DeleteRequest:
        id: int


Model.LoadAllResponse.__pydantic_model__.update_forward_refs()


@dataclass(frozen=True)
class TestCaseInfo:
    id: int
    name: str
    sample_count: int
    # provided by stratification preview to show what bucket this test-case falls into,
    # e.g. {"count": "low", "gender": "female"}. This info is not available from stratification result.
    membership: Optional[Dict[str, str]] = None


class BulkProcessStatus(Enum):
    CREATED = "created"
    LOADED = "loaded"
    EDITED = "edited"


class TestCase:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str
        workflow: str

    @dataclass(frozen=True)
    class SingleProcessRequest:
        name: str
        reset: bool = False

    @dataclass(frozen=True)
    class SingleProcessResponse:
        data: "TestCase.EntityData"
        status: BulkProcessStatus

    @dataclass(frozen=True)
    class BulkProcessRequest:
        test_cases: List["TestCase.SingleProcessRequest"]
        workflow: str
        uuid: Optional[str] = None

    @dataclass(frozen=True)
    class BulkProcessResponse:
        test_cases: List["TestCase.SingleProcessResponse"]

    @dataclass(frozen=True)
    class CreateFromExistingRequest(BatchedLoad.WithLoadUUID):
        test_case_name: str
        test_suite_name: str
        workflow: str
        compute_metrics_where_possible: bool = False
        compute_metrics_models: Optional[List[int]] = None

    @dataclass(frozen=True)
    class CreateFromExistingResponse:
        test_case_id: int
        test_suite_id: int  # newly created version ID

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        workflow: str

    @dataclass(frozen=True)
    class LoadContentsRequest:
        test_case_id: int

    @dataclass(frozen=True)
    class InitLoadContentsRequest(LoadContentsRequest, BatchedLoad.BaseInitDownloadRequest):
        ...

    @dataclass(frozen=True)
    class EditRequest:
        test_case_id: int
        current_version: int
        description: str
        reset: bool = False

    @dataclass(frozen=True)
    class CompleteEditRequest(EditRequest, BatchedLoad.WithLoadUUID):
        ...


TestCase.SingleProcessResponse.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]
TestCase.BulkProcessRequest.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]
TestCase.BulkProcessResponse.__pydantic_model__.update_forward_refs()  # type: ignore[attr-defined]


class TestSuite:
    @dataclass(frozen=True)
    class CreateRequest:
        name: str
        description: str
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class LoadByNameRequest:
        name: str
        version: Optional[int] = None

    @dataclass(frozen=True)
    class LoadAllRequest:
        workflow: str
        tags: Optional[List[str]] = None

    @dataclass(frozen=True)
    class EntityData:
        id: int
        name: str
        version: int
        description: str
        test_cases: List[TestCase.EntityData]
        workflow: str
        tags: List[str]

    @dataclass(frozen=True)
    class LoadAllResponse:
        test_suites: List["TestSuite.EntityData"]

    @dataclass(frozen=True)
    class EditRequest:
        test_suite_id: int  # ID of version being edited
        current_version: int
        name: str
        description: str
        test_case_ids: List[int]
        tags: Optional[List[str]] = None  # unique set -- list used to preserve ordering

    @dataclass(frozen=True)
    class DeleteRequest:
        test_suite_id: int


TestSuite.LoadAllResponse.__pydantic_model__.update_forward_refs()


class TestRun:
    @dataclass(frozen=True)
    class MarkCrashedRequest:
        test_run_id: int
